# Copyright (C) 2009, Olli Sikstus
# License: Creative Commons Attribution-Share Alike 3.0 Unported
# http://creativecommons.org/licenses/by-sa/3.0/

import math
import random
import pygame

import gfxman

class Map:
	def __init__(self, pixelWidth, pixelHeight):
		self.tileMapW = int(pixelWidth / 64)
		self.tileMapH = int(pixelHeight / 64)
		self.lightMapW = self.tileMapW * 2
		self.lightMapH = self.tileMapH * 2
		self.pixelW = self.tileMapW * 64
		self.pixelH = self.tileMapH * 64
		
		self.tileMap = [0] * (self.tileMapW * self.tileMapH)
		for i in xrange(self.tileMapW * self.tileMapH):
			if random.randint(1, 15) == 1:
				self.tileMap[i] = 1
			if random.randint(1, 10) == 1:
				self.tileMap[i] = 2
		self.lightMap = [0] * (self.lightMapW * self.lightMapH)
	
	def clearLightMap(self):
		del self.lightMap
		self.lightMap = [0] * (self.lightMapW * self.lightMapH)
	
	# Cast light from (lightmap) tile
	def castLightFromTile(self, x, y, intensity, dist):
		if x >= 0 and x < self.lightMapW and y >= 0 and y < self.lightMapH:
			self.setLight(x, y, intensity)
		for angle in xrange(1, 360, 2):
			rad = math.radians(angle)
			light = intensity
			curX = x
			curY = y
			cos = math.cos(rad)
			sin = math.sin(rad)
			block = 0
			for i in xrange(dist):
				# Thanks for the tutorial, Sol
				v = float(i) / float(dist)
				v = v * v
				light = int(255.0 * (1 - v)) - block
				if light >= 0:
					curX += cos
					curY += sin
					iX = int(curX)
					iY = int(curY)
					if iX >= 0 and iX < self.lightMapW and iY >= 0 and iY < self.lightMapH:
						if self.getLight(iX, iY) < light:
							self.setLight(iX, iY, light)
						if self.getTileMapTile(iX / 2, iY / 2) == 2:
							block += 96
				else:
					break
	
	def getLight(self, x, y):
		return self.lightMap[y * self.lightMapW + x]
	
	def setLight(self, x, y, light):
		self.lightMap[y * self.lightMapW + x] = light
	
	def getTileMapTile(self, x, y):
		return self.tileMap[y * self.tileMapW + x]
	
	def safeGetTileMapTile(self, x, y):
		if x >= 0 and x < self.tileMapW and y >= 0 and y < self.tileMapH:
			return self.getTileMapTile(x, y)
		else:
			return -1
	
	def drawTiles(self, gfx, cameraX, cameraY):
		for y in xrange(self.tileMapH):
			for x in xrange(self.tileMapW):
				screenX = x * 64 - cameraX
				screenY = y * 64 - cameraY
				if screenX >= -64 and screenX < 800 and screenY >= -64 and screenY < 600:
					tile = self.getTileMapTile(x, y)
					if tile == 0:
						gfx.draw("grass.png", screenX, screenY)
					elif tile == 1:
						gfx.draw("grass2.png", screenX, screenY)
					elif tile == 2:
						gfx.draw("tree.png", screenX, screenY)
	
	def drawLights(self, gfx, cameraX, cameraY):
		for y in xrange(self.lightMapH):
			for x in xrange(self.lightMapW):
				screenX = x * 32 - cameraX
				screenY = y * 32 - cameraY
				if screenX >= -32 and screenX < 800 and screenY >= -32 and screenY < 600:
					light = self.getLight(x, y)
					if light < 255:
						gfx.drawAlpha("darkness.png", 255 - light, screenX, screenY)
	
	def canMovePoint(self, pX, pY, tX, tY, dist):
		t = self.safeGetTileMapTile(tX, tY)
		if t == 2:
			if ((pX - ((tX * 64) + 32))**2 + (pY - ((tY * 64) + 32))**2) < (dist**2) * 2:
				return False
		elif t == -1:
			return False
		return True
	
	def canMoveCircle(self, fX, fY):
		pX = int(fX)
		pY = int(fY)
		if pX < 16 or pY < 16:
			return False
		if pX >= self.pixelW - 16 or pY >= self.pixelH - 16:
			return False
		tX = (pX - 32) / 64
		tY = (pY - 32) / 64
		if self.canMovePoint(pX, pY, tX, tY, 32) == False:
			return False
		tX = (pX - 32) / 64
		tY = (pY + 32) / 64
		if self.canMovePoint(pX, pY, tX, tY, 32) == False:
			return False
		tX = (pX + 32) / 64
		tY = (pY - 32) / 64
		if self.canMovePoint(pX, pY, tX, tY, 32) == False:
			return False
		tX = (pX + 32) / 64
		tY = (pY + 32) / 64
		if self.canMovePoint(pX, pY, tX, tY, 32) == False:
			return False
		return True
	
	def canMove(self, creatureX, creatureY):
		if self.canMoveCircle(creatureX + 32, creatureY + 32) == False:
			return False
		return True

class Game:
	def __init__(self, screenSize = (800, 600), fullScreen = False, title = ""):
		pygame.init()
		self.gfx = gfxman.RabbytGFX(screenSize, fullScreen, title)
		pygame.display.set_caption(title)
		
		self.shootWav = pygame.mixer.Sound("shoot.wav")
		pygame.mixer.music.load("music.ogg")
		
		self.cameraX = 0
		self.cameraY = 0
		self.fader = 0
		self.die = False
		self.gameOver = False
		self.end = False
		
		self.map = Map(800 * 2, 600 * 2)
		self.playerX = random.randint(0, self.map.pixelW - 64 - 1)
		self.playerY = random.randint(0, self.map.pixelH - 64 - 1)
		while self.map.canMove(self.playerX, self.playerY) == False:
			self.playerX = random.randint(0, self.map.pixelW - 64 - 1)
			self.playerY = random.randint(0, self.map.pixelH - 64 - 1)
		self.playerRot = random.randint(0, 359)
		self.playerTargeting = False
		self.shootX = None
		self.shootY = None
		
		self.monsterX = random.randint(0, self.map.pixelW - 64 - 1)
		self.monsterY = random.randint(0, self.map.pixelH - 64 - 1)
		while self.map.canMove(self.monsterX, self.monsterY) == False:
			self.monsterX = random.randint(0, self.map.pixelW - 64 - 1)
			self.monsterY = random.randint(0, self.map.pixelH - 64 - 1)
		self.monsterRot = random.randint(0, 359)
		self.monsterState = "wander"
		self.monsterSpeed = 1.0
		self.monsterDir = 1
		
		self.oldLightX = -1
		self.oldLightY = -1
		
		pygame.mixer.music.play(-1)
	
	def isRunning(self):
		return not self.end
	
	def centerCamera(self, cX, cY):
		self.cameraX = int(cX) - 400
		self.cameraY = int(cY) - 300
		if self.cameraX < 0:
			self.cameraX = 0
		if self.cameraX >= self.map.pixelW - 800:
			self.cameraX = self.map.pixelW - 800 - 1
		if self.cameraY < 0:
			self.cameraY = 0
		if self.cameraY >= self.map.pixelH - 600:
			self.cameraY = self.map.pixelH - 600 - 1
	
	def shoot(self):
		rad = math.radians(-self.playerRot - 90)
		curX = self.playerX + 32
		curY = self.playerY + 32
		cos = math.cos(rad)
		sin = math.sin(rad)
		while True:
			curX += cos
			curY += sin
			iX = int(curX)
			iY = int(curY)
			if ((iX - (self.monsterX + 32))**2 + (iY - (self.monsterY + 32))**2) < 32**2:
				self.gameOver = True
				return iX, iY
			if self.map.canMovePoint(iX, iY, iX / 64, iY / 64, 16) == False:
				return iX, iY
	
	def movePlayer(self, dT):
		key = pygame.key.get_pressed()
		keyMods = pygame.key.get_mods()
		if keyMods & pygame.KMOD_LSHIFT:
			if self.playerTargeting == False:
				self.playerTargetCounter = 1000
			self.playerTargeting = True
		else:
			self.playerTargetCounter = 1000
			self.playerTargeting = False
		if self.playerTargeting == True:
			if self.playerTargetCounter > 0:
				self.playerTargetCounter -= dT
				if self.playerTargetCounter < 0:
					self.playerTargetCounter = 0
		
		if self.playerTargeting == False:
			if key[pygame.K_a]:
				if self.playerX > 0:
					pX = self.playerX - 0.1 * dT
					if self.map.canMove(pX, self.playerY):
						self.playerX = pX
			if key[pygame.K_d]:
				if self.playerX < self.map.pixelW - 64:
					pX = self.playerX + 0.1 * dT
					if self.map.canMove(pX, self.playerY):
						self.playerX = pX
			if key[pygame.K_w]:
				if self.playerY > 0:
					pY = self.playerY - 0.1 * dT
					if self.map.canMove(self.playerX, pY):
						self.playerY = pY
			if key[pygame.K_s]:
				if self.playerY < self.map.pixelH - 64:
					pY = self.playerY + 0.1 * dT
					if self.map.canMove(self.playerX, pY):
						self.playerY = pY
		if ((self.playerX - self.monsterX)**2 + (self.playerY - self.monsterY)**2) < (32**2) * 2:
			self.gameOver = True
			self.die = True
		if self.playerTargeting == True and self.playerTargetCounter == 0:
			b1, b2, b3 = pygame.mouse.get_pressed()
			if b1 == True:
				self.playerTargetCounter = 1000
				self.shootWav.play()
				self.shootX, self.shootY = self.shoot()
		
		mouseX, mouseY = pygame.mouse.get_pos()
		mouseX += self.cameraX
		mouseY += self.cameraY
		self.playerRot = math.degrees(math.atan2(self.playerX - mouseX, self.playerY - mouseY))
		
		lX = int(self.playerX + 32.0 + (11.0 * math.cos(math.radians(-self.playerRot - 90 - 45))))
		lY = int(self.playerY + 32.0 + (11.0 * math.sin(math.radians(-self.playerRot - 90 - 45))))
		self.lightX = lX / 32
		self.lightY = lY / 32
		if self.lightX != self.oldLightX or self.lightY != self.oldLightY:
			self.oldLightX = self.lightX
			self.oldLightY = self.lightY
			self.map.clearLightMap()
			self.map.castLightFromTile(self.lightX, self.lightY, 255, 10)
	
	def monsterAi(self, dT):
		dist2 = ((self.playerX - self.monsterX)**2 + (self.playerY - self.monsterY)**2)
		if dist2 < (96**2) * 2:
			self.monsterState = "attack"
			self.monsterSpeed = 2.75
		elif dist2 < (256**2) * 2:
			if self.monsterState == "attack":
				if dist2 > (160**2) * 2:
					self.monsterState = "escape"
					self.monsterSpeed = 2.4
			else:
				self.monsterState = "escape"
				self.monsterSpeed = 2.4
		else:
			if self.monsterState == "attack" or self.monsterState == "escape":
				self.monsterState = "wander"
				self.monsterSpeed = 1.75
		
		if self.monsterState == "attack":
			self.monsterRot = math.degrees(math.atan2(self.monsterX - self.playerX, self.monsterY - self.playerY))
		elif self.monsterState == "escape":
			self.monsterRot = math.degrees(math.atan2(self.monsterX - self.playerX, self.monsterY - self.playerY)) + 180
			if self.monsterRot >= 360:
				self.monsterRot -= 360
		elif self.monsterState == "wander":
			if random.randint(1, 200) == 1:
				self.monsterRot = random.randint(0, 359)
		if random.randint(1, 200) == 1:
			self.monsterDir *= -1
		while True:
			mX = self.monsterX + (self.monsterSpeed * math.cos(math.radians(-self.monsterRot - 90)))
			mY = self.monsterY + (self.monsterSpeed * math.sin(math.radians(-self.monsterRot - 90)))
			if self.map.canMove(mX, mY):
				self.monsterX = mX
				self.monsterY = mY
				break
			else:
				self.monsterRot -= 2 * self.monsterDir
				if self.monsterRot < 0:
					self.monsterRot += 360
				elif self.monsterRot >= 360:
					self.monsterRot -= 360
	
	def gameLoop(self, dT):
		self.movePlayer(dT)
		self.monsterAi(dT)
		
		self.centerCamera(self.playerX + 32, self.playerY + 32)
		
		self.gfx.beginFrame()
		self.map.drawTiles(self.gfx, self.cameraX, self.cameraY)
		playerSprite = "player.png"
		if self.playerTargeting == True:
			playerSprite = "player_gun.png"
		self.gfx.drawRot(playerSprite, self.playerRot, int(self.playerX) - self.cameraX, int(self.playerY) - self.cameraY)
		self.gfx.drawRot("monster.png", self.monsterRot, int(self.monsterX) - self.cameraX, int(self.monsterY) - self.cameraY)
		self.map.drawLights(self.gfx, self.cameraX, self.cameraY)
		if self.playerTargeting == True and self.playerTargetCounter > 0:
			a = int((float(self.playerTargetCounter) / 1000.0) * 255.0)
			self.gfx.draw("targeting_green.png", 10, 10)
			self.gfx.drawAlpha("targeting_red.png", a, 10, 10)
		if self.shootX and self.shootY:
			self.gfx.draw("shot.png", self.shootX - 4 - self.cameraX, self.shootY - 4 - self.cameraY)
			self.shootX = None
			self.shootY = None
		self.gfx.endFrame()
	
	def loop(self):
		# Handle events
		for event in pygame.event.get():
			if event.type == pygame.QUIT or (event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE):
				self.end = True
		
		dT = self.gfx.getTimeDelta()
		
		pygame.event.pump()
		
		if not self.gameOver:
			self.gameLoop(dT)
		else:
			if self.fader == 0:
				pygame.mixer.music.fadeout(2500)
			if self.die == True:
				if self.fader <= 200:
					r = float(self.fader) / 200.0
					self.gfx.clearTo(r, 0.0, 0.0)
				elif self.fader <= 2200:
					r = float(self.fader - 200) / 2000.0
					r = 1.0 - r
					self.gfx.clearTo(r, 0.0, 0.0)
				else:
					self.end = True
				self.fader += dT
				self.gfx.endFrame()
			else:
				if self.fader <= 200:
					c = float(self.fader) / 200.0
					self.gfx.clearTo(c, c, c)
				elif self.fader <= 2200:
					c = float(self.fader - 200) / 2000.0
					c = 1.0 - c
					self.gfx.clearTo(c, c, c)
				else:
					self.end = True
				self.fader += dT
				self.gfx.endFrame()

def main():
	game = Game(title = "Missing Link")
	
	while game.isRunning():
		game.loop()
	
	print "Fps: %.2f" % game.gfx.getFps()

if __name__ == "__main__":
	main()
